/*PROGRAM mqget.c */
/*
 *  Retrieve message from an MQ Server and place them into unique files
 *
 *
 * 14-05-2001  sidy  Errors reporting from mqget
 *                   doubled malloc request and added messages.
 * 29-10-2001        Included print of Application ID (filename) from Queue.
 * 30-11-2006 BruceH Added MQGMO_LOCK option to prevent losing some &
 *                   duplicating other data files
 * 05-02-2007        Modified for AES support
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <cmqc.h>
#include "rijndael.h"

time_t  longtime;
struct tm *timenow;

unsigned int rfFlag;	/* flag for "real filename" */
unsigned int dhFlag;	/* flag for "dump Header" */
unsigned int aesFlag;	/* flag for aes support ON */

char *_mqBuffPtr;
char key[32];

char szBuffer[1024];
char szQueueManager[64];
char szFileName[64];
char szAESKey[64];
char szQueueName[64];


/*
 * Declare MQI structures needed
 */
MQOD  od = {MQOD_DEFAULT};/* Object Descriptor */
MQMD  md = {MQMD_DEFAULT};/* Message Descriptor*/
MQGMO gmo= {MQGMO_DEFAULT};   /* get message options   */


/*
 * header block used for AES/Rijndael encrypted messages
 *
 */
struct HDR_BLOCK
{
	unsigned char hdr_length[4];	/* platform neutral values */
	unsigned char hdr_major;
	unsigned char hdr_minor;
	char src_hash[34];
	unsigned char file_length[4];
	char file_name[64];
	char file_type[4];
	unsigned char hdr_flag_1;	/* flag used to hold 'E'ncrypted data block */
	unsigned char hdr_flag_2;
	unsigned char hdr_flag_3;
	unsigned char hdr_flag_4;	/* OS Flag for PUT utility */
};






/**
 * getHeaderLength()
 *
 *
 * Utility to return length of header
 * given pointer to header block
 *
 *
 */
long getHeaderLength( struct HDR_BLOCK *hdr )
{
static long _HdrBlockLength;

	_HdrBlockLength =hdr->hdr_length[0];
	_HdrBlockLength|=hdr->hdr_length[1]<<8;
	_HdrBlockLength|=hdr->hdr_length[2]<<16;
	_HdrBlockLength|=hdr->hdr_length[3]<<24;
	return _HdrBlockLength;
}








/**
 * getFileLength()
 *
 *
 * Helper utility to return file length stored in 
 * header block passed in message buffer.
 *
 *
 *
 *
 */
long getFileLength( struct HDR_BLOCK *hdr )
{
static long _SavedFileLength;

	_SavedFileLength= hdr->file_length[0];
	_SavedFileLength|=hdr->file_length[1]<<8;
	_SavedFileLength|=hdr->file_length[2]<<16;
	_SavedFileLength|=hdr->file_length[3]<<24;
	return _SavedFileLength;
}







/**
 * GetNextSeq()
 *
 *
 * Returns the next sequence number to use when -rf is NOT specified
 *
 *
 * @pre    Sequence file exists in /tmp
 * @post   Sequence file content incremented by 1
 * @return long
 */
long GetNextSeq( char *szNumber)
{
FILE * seqfPtr;
long seqNumber=0L;

	sprintf(szNumber,"%0ld", seqNumber);
	if((seqfPtr=fopen("/tmp/.mqsequence", "r+")) != NULL)
	{
		fgets(szBuffer,80,seqfPtr);
		seqNumber = atol(szBuffer);
		rewind(seqfPtr);
		fprintf(seqfPtr, "%ld", ++seqNumber);
		fclose(seqfPtr);
		sprintf(szNumber,"%0ld", seqNumber);
		printf("Sequence number:  %0ld\n", seqNumber);
		return 0;
	}
	else
	{
		if((seqfPtr = fopen("/tmp/.mqsequence","w+"))!= NULL)
		{
			fprintf(seqfPtr,"0000\n");
			fclose(seqfPtr);
			chmod("/tmp/.mqsequence", 0666);
			return 0;
		}
		else
		{
			printf("ERROR - Unable to open /tmp/.mqsequence - Please create manually\n\n");
			return 1;
		}
	}
	return 1;
}







/**
 * SaveMessage2File()
 *
 *
 *
 * Given the specific filename save data to this file.
 *
 * @pre    Valid ptr to data and valid filename
 * @post   File saved to disk else error
 * @return void
 */
void SaveMessage2File( char *ptr, MQLONG lBufferSize, char *file )
{
FILE *fPtr;
	
	printf("Saving MessageFile to %s\n", file);
	if((fPtr=fopen(file,"w+"))!=NULL)
	{
		while(lBufferSize--)
			fputc( *ptr++, fPtr);
	}
	else
	{
		fprintf(stderr,"ERROR - Cannot create output file (%s)\n",file);
		return;
	}
	fclose(fPtr);
}




/**
 * SaveMessage()
 *
 *
 *
 * Create a new file based on the sequence number
 *
 * @preValid ptr to data 
 * @post   File saved to disk else error
 * @return void
 */
void SaveMessage( char *ptr, MQLONG lBufferSize )
{
char szBuffer[256];
FILE *fPtr;
char seqNumber[16];

	if(dhFlag==1) printf("SaveMessage()\n");
	if(GetNextSeq(seqNumber)!=0)
	{
		printf("ERROR - Cannot allocate sequence number\n");
		printf("ERROR - File save aborted\n");
		return;
	}
	time(&longtime);
	timenow=localtime( &longtime );
	sprintf(szBuffer, "mq%04d%02d%02d.%s", 
		timenow->tm_year+1900,
		timenow->tm_mon+1,
		timenow->tm_mday,
		seqNumber);
	printf("Saving Message to %s\n", szBuffer);
	if((fPtr=fopen(szBuffer,"w+"))!=NULL)
	{
		while(lBufferSize--)
			fputc( *ptr++, fPtr);
	}
	else
	{
		fprintf(stderr,"ERROR - Cannot create output file (%s)\n",szBuffer);
		return;
	}
	fclose(fPtr);
}






/**
 * DumpBlock()
 *
 *
 * Dump the Header to screen if the -dh flag
 * is specified on the command line
 *
 * @pre    None
 * @post   None
 * @return void
 */
void DumpBlock(struct HDR_BLOCK *hdr)
{
	if(dhFlag==0) return;

	printf("Header Dump\n");
	printf("===========\n");
	printf("Version    [%c.%c] \n", hdr->hdr_major, hdr->hdr_minor);
	printf("Block Size [%ld]\n", getHeaderLength(hdr));
	printf("Hash       [%s]\n", hdr->src_hash);
	printf("File Name  [%s]\n", hdr->file_name);
	printf("File Length[%ld]\n", getFileLength(hdr));
	printf("FileType   [%s]\n", hdr->file_type);
	printf("Flag 1     [%c]\n", hdr->hdr_flag_1);
	printf("Flag 2     [%c]\n", hdr->hdr_flag_2);
	printf("Flag 3     [%c]\n", hdr->hdr_flag_3);
	printf("Flag 4     [%c]\n", hdr->hdr_flag_4);
}




/*
 * DecryptData()
 *
 * 
 * Common method to do a block by block decryption of the message buffer
 * returned via the MQGET call.
 * Buffer is 2 times the message size plus 32 bytes extra.
 * 
 * @Pre    Buffer pSource contains AES data in 16 char blocks
 * @Post   Buffer pDest is filled with clear text
 * @return void 
 */
void DecryptData( struct HDR_BLOCK * hdr, rijndael_ctx *ctx, char *pSource, char *pDest)
{
long flen;		/* original file length */
long blkcnt;	/* number of 16 byte blocks */
long maxblks;	/* max blocks to fit all data */
long maxflen;	/* size in bytes */
long blklen;	/* length of header block */

 	if(dhFlag==1) printf("DecryptData()\n");

	DumpBlock( hdr );

	blklen=getHeaderLength(hdr);
	pSource+=blklen;			/* start of AES data + header offset*/
	flen=getFileLength(hdr);	/* original file length from header */
	blkcnt=flen/16;		/* number of 16 byte blocks */
	maxblks=blkcnt;		/* max blocks to fit all data */
	maxflen=maxblks*16;	/* size in bytes */
	printf("BLKS [%ld] LEN [%ld] MAX [%ld]\n", maxblks, flen, maxflen);
	if((blkcnt*16)!=flen)
	{
		if(dhFlag==1) printf("Adjusting blocks\n");
		maxblks++; maxflen=maxblks*16;
	}
	if(dhFlag==1) printf(" <set_key> ");
	rijndael_set_key(ctx, key, 128, 0);
	while(maxblks--)
	{
		if(dhFlag==1) printf(".");
		rijndael_decrypt(ctx, pSource, pDest);
		pSource+=16; pDest+=16;
	}
}







/*=================================================================
 *
 *
 *
 *
 *                      M A I N
 *
 *
 *
 *
 *
 */
int main(int argc, char **argv)
{
MQHCONN hConn;			/* connection handle*/
MQHOBJ  hObject;		/* object handle*/
MQLONG _OpenOptions;	/* MQOPEN options*/
MQLONG _CloseOptions;	/* MQCLOSE options*/
MQLONG _CC;				/* completion code*/
MQLONG _OpenCode;		/* MQOPEN completion code*/
MQLONG _RC;
MQLONG _ConnRC;		/* reason code for MQCONN*/
MQBYTE buffer[101];	/* message buffer*/
MQLONG buflen;		/* buffer length*/
MQLONG buffLength;		/* default length */
MQLONG _msgLength;		/* message length received*/
MQLONG exitCode;
int i;
long blklen;

struct HDR_BLOCK *hdr;
rijndael_ctx *ctx;

	rfFlag=0;
	dhFlag=0;
	aesFlag=0;
	time(&longtime);
	srandom(longtime);
	memset(szAESKey,0,sizeof(szAESKey));
	memset(szQueueName,0,sizeof(szQueueName));
	memset(szQueueName,0,sizeof(szQueueName));
	memset(szQueueManager,0,sizeof(szQueueManager));
	memset(szFileName,0,sizeof(szFileName));
	
	printf("\nMQGET v2.0 (aes) - Retrieve ALL messages from a Queue\n");
	printf("Build Date: %s  Time: %s\n", __DATE__, __TIME__);
	printf("Author: Sid Young\n");
	if(argc < 2)
	{
		printf("Minimum required parameter missing - <queue name>\n");
		exit(99);
	}
/*
 *---------------------------------------------
 *       Process command line args
 *---------------------------------------------
 */
	if(argc>1)
	{
		for(i=0;i<argc;++i)
		{
			if(!strcmp(argv[i],"-?"))
			{
				printf("HELP: mqget <queue_name> [-key <key>] [-rf -dh]\n");
				printf(" <queue_name> - Name of Queue to get from\n");
				printf(" -key <key> - 16 byte hex string for cipher\n");
				printf(" -rf - Retrieve file name if available\n");
				printf(" -dh - dump header if encrypted\n");
				exit(0);
			}
			if(!strcmp(argv[i],"-rf")) { rfFlag=1; }
			if(!strcmp(argv[i],"-dh")) { dhFlag=1; }
			if(!strcmp(argv[i],"-key")) {aesFlag=1;strcpy(szAESKey,argv[i+1]);}
			if(!strcmp(argv[i],"-int"))
			{
				printf("Sizeof int = %d\n", sizeof(unsigned int));
				printf("Sizeof long= %ld\n", sizeof(unsigned long));
			}
		}
		if(dhFlag==1)
		{
			printf("Command line options:\n");
			for(i=0;i<argc;++i)
	 			printf("Arg [%d] Value [%s]\n",i, argv[i]);
		}
	}
	strcpy(od.ObjectName, argv[1]);
	strcpy(szQueueName,argv[1]);
	szQueueManager[0] = 0;   /* just use default */
/*
 *---------------------------------------------
 * Allocate a header to hold details about message
 *---------------------------------------------
 */
 	if(aesFlag==1)
	{
		printf("AES decryption required\n");
  		if((hdr=(struct HDR_BLOCK *)malloc(sizeof(struct HDR_BLOCK)))==NULL)
		{
			printf("ERROR - Cannot allocate HDR block\n");
			exit(-1);
		}
  		if((ctx=(rijndael_ctx *)malloc(sizeof(rijndael_ctx)))==NULL)
		{
			printf("ERROR - Cannot allocate CTX block\n");
			exit(-1);
		}
		memset(hdr,0,sizeof(struct HDR_BLOCK));
/*
 *---------------------------------------------
 * make sure key is correct length 128 bits minimum
 *---------------------------------------------
 */
		while(strlen(szAESKey)<32)
			strcat(szAESKey, "0");
		memcpy(key,szAESKey,32);
	}
/*
 *---------------------------------------------
 * Connect to queue manager
 *---------------------------------------------
 */
	printf("Attempt Connect ->  ");
	MQCONN(szQueueManager, &hConn, &_CC, &_ConnRC);
/*
 *---------------------------------------------
 * report reason and stop if it failed
 *---------------------------------------------
 */
	if(_CC == MQCC_FAILED)
	{
		printf("ERROR - MQCONN ended with RC [%ld]\n", _ConnRC);
	   	exit( 98 );
	}
	printf("Connected to Queue Manager\n");
/*
 *---------------------------------------------
 * Open the named message queue for input; exclusive or shared
 * use of the queue is controlled by the queue definition here
 *---------------------------------------------
 */
	_OpenOptions = MQOO_INPUT_AS_Q_DEF+MQOO_FAIL_IF_QUIESCING+MQOO_BROWSE;
	printf("Open queue....\n");
	MQOPEN(hConn, &od, _OpenOptions, &hObject, &_OpenCode, &_RC);
/* 
 *---------------------------------------------
 *  Report reason, if any; stop if failed
 *---------------------------------------------
 */
	if(_RC != MQRC_NONE)
	{
		printf("MQOPEN ended with reason code [%ld]\n", _RC);
	}
	if(_OpenCode == MQCC_FAILED)
	{
		printf("unable to open queue for input\n");
	}
/*
 *---------------------------------------------
 * Get messages from the message queue
 * Loop until there is a failure
 *---------------------------------------------
 */
	_CC = _OpenCode;   /* use MQOPEN result for initial test  */
	gmo.Version = MQGMO_VERSION_2; /* Avoid need to reset Message */
	gmo.MatchOptions = MQMO_NONE;  /* ID and Correlation ID after */
	gmo.WaitInterval = 100;
	gmo.Options=MQGMO_WAIT+MQGMO_LOCK+MQGMO_ACCEPT_TRUNCATED_MSG+MQGMO_BROWSE_FIRST;

	md.Encoding   = MQENC_NATIVE;
	md.CodedCharSetId = MQCCSI_Q_MGR;
/*
 *---------------------------------------------
 * Sit in loop and collect messages from queue
 *---------------------------------------------
 */
	exitCode=0;
	while(_CC != MQCC_FAILED)
	{
		printf("Browse queue [%s] for messages\n", szQueueName);
/*
 *  Get size of msg for malloc - non destructive browse
 */
		MQGET(hConn,hObject,&md,&gmo,0L,_mqBuffPtr,&_msgLength,&_CC,&_RC);
		if(_RC==2033) _msgLength=0;
		printf("Msg Len [%ld]  RC [%ld]  CC [%ld]\n", _msgLength, _RC,_CC);
/*
 * Possible reason codes:
 * 2079 - truncated msg returned
 * 2080 - Truncated message failed
 * 2033 - No more messages.
 */
		if(_RC == MQRC_NO_MSG_AVAILABLE)
		{
			printf("No messages available\n");
			break;		/* exit loop here */
		}
/*
 *---------------------------------------------
 * Allocate message buffer
 * use last call to get size
 * this call to get data into buffer.
 *---------------------------------------------
 */
		gmo.Options = MQGMO_WAIT+MQGMO_BROWSE_MSG_UNDER_CURSOR+MQGMO_LOCK;
		if((_mqBuffPtr=malloc((_msgLength*2)+32))!=NULL)
		{
			char *dst= _mqBuffPtr+_msgLength+1;
			char *pOrgDest= _mqBuffPtr+_msgLength+1;
			char *org_mqBuffPtr= _mqBuffPtr; /* save for free() done later */
/*
 *---------------------------------------------
 *  Get actual Message into buffer
 *---------------------------------------------
 */
 			printf("Retrieve message from queue\n");
			MQGET(hConn,hObject,&md,&gmo,_msgLength,_mqBuffPtr,&_msgLength, &_CC, &_RC);
			buffLength = _msgLength;		/* default length*/
			if((rfFlag==1)&&(strlen(md.ApplIdentityData)>1))
			{
				memset( szFileName,0,sizeof(szFileName));
				mqTrim(32,md.ApplIdentityData, szFileName, &_CC, &_RC);
				printf("RAW file name [%s]\n", szFileName);
				for(i=0;i<32;i++)
				{
					if(szFileName[i]==' ') { szFileName[i]=0; break; }
				}
			}
 			printf("Retrieve message from queue\n");
			if(aesFlag==1)
			{
				printf("Decrypting data ->");
				memcpy(hdr, _mqBuffPtr, 4); /* first 4 bytes are blk size*/
				blklen=getHeaderLength(hdr);
				printf("BLKLEN [%d]  0=%02X  1=%02X\n", blklen, hdr[0], hdr[1]);
				memcpy(hdr,_mqBuffPtr,blklen);		/* mv header in message */

				DecryptData( hdr, ctx, _mqBuffPtr, dst );
				buffLength =getFileLength(hdr);		/* use file size from HDR*/
				printf(" Done!\n");
			}
			else
			{
				printf("Data is cleartext\n");
				pOrgDest=org_mqBuffPtr;
			}
/*
 *---------------------------------------------
 * Save file to disk, either with returned name
 * or use constructed numerically unique name.
 *---------------------------------------------
 */
			if((rfFlag==1) && (strlen(md.ApplIdentityData)>1))
				SaveMessage2File( pOrgDest, buffLength, szFileName );
			else
				SaveMessage( pOrgDest, buffLength );

			exitCode=1;
			gmo.Options=MQGMO_WAIT+MQGMO_COMPLETE_MSG+MQGMO_ACCEPT_TRUNCATED_MSG;
			MQGET(hConn,hObject,&md,&gmo,_msgLength,_mqBuffPtr,&_msgLength, &_CC, &_RC);
			free(org_mqBuffPtr);
		}
		else
		{
			printf("ERROR - Cannot allocate buffer for GET\n");
			exit(-1);
		}
		gmo.Options=MQGMO_WAIT+MQGMO_BROWSE_NEXT+MQGMO_ACCEPT_TRUNCATED_MSG+MQGMO_LOCK;
		_mqBuffPtr = szBuffer;	/* reset bufer to something */
		md.Encoding = MQENC_NATIVE;
		md.CodedCharSetId = MQCCSI_Q_MGR;

		printf("======= Next ========\n");

	} // end of while loop

/*
 *---------------------------------------------
 * Close the source queue (if it was opened)
 *---------------------------------------------
 */
   if(_OpenCode != MQCC_FAILED)
   {
		_CloseOptions = 0;   /* no close options*/
		MQCLOSE(hConn, &hObject, _CloseOptions, &_CC, &_RC);
		if(_RC != MQRC_NONE)
		{
			printf("MQCLOSE ended with reason code %ld\n", _RC);
		}
	}
/*---------------------------------------------
 * Disconnect from MQM if not already connected
 *---------------------------------------------
 */
	if(_ConnRC != MQRC_ALREADY_CONNECTED )
	{
		MQDISC(&hConn, &_CC, &_RC); 
		if(_RC != MQRC_NONE)
		{
			printf("MQDISC ended with reason code %ld\n", _RC);
		}
	}

	printf("MQGET ended. RC=%d\n", _RC);
	exit((int)exitCode);
}
/* end of file */
